@@ -113,6 +113,7 @@ YFrameClient::~YFrameClient() {
     if (fSizeHints) { XFree(fSizeHints); fSizeHints = nullptr; }
     if (fHints) { XFree(fHints); fHints = nullptr; }
     if (fClientItem) { fClientItem->goodbye(); fClientItem = nullptr; }
+    if (fPinging) { fPingTimer = null; }
 }
 
 void YFrameClient::getProtocols(bool force) {
@@ -457,11 +458,18 @@ bool YFrameClient::handleTimer(YTimer* timer) {
             }
             else if (fPid > 0) {
                 char* res = classHint()->resource();
+                if (res && strlen(res) > 47L) {
+                    char* dot = strchr(res, '.');
+                    if (dot && inrange<long>(dot - res, 8L, 48L))
+                        *dot = '\0';
+                    else
+                        res[47] = '\0';
+                }
                 char buf[234];
                 snprintf(buf, sizeof buf,
                          _("Client %s with PID %ld fails to respond.\n"
                            "Do you wish to terminate this client?\n"),
-                         res, fPid);
+                         res ? res : "''", fPid);
                 fFrame->wmConfirmKill(buf);
                 free(res);
             }
@@ -470,6 +478,14 @@ bool YFrameClient::handleTimer(YTimer* timer) {
             }
         }
     }
+    if (fUrgencyTimer == timer) {
+        fUrgencyTimer = null;
+        xsmart<XWMHints> h(XGetWMHints(xapp->display(), handle()));
+        if (h && hasbit(h->flags, XUrgencyHint)) {
+            h->flags &= ~XUrgencyHint;
+            XSetWMHints(xapp->display(), handle(), h);
+        }
+    }
 
     return false;
 }
@@ -630,11 +646,17 @@ void YFrameClient::handleProperty(const XPropertyEvent &property) {
     case XA_WM_CLASS:
         prop.wm_class = new_prop;
         if (prop.wm_class) {
-            ClassHint old(fClassHint);
+            ClassHint previous(fClassHint);
             getClassHint();
-            if (fClassHint.nonempty() && fClassHint != old) {
+            if (fClassHint.nonempty() && fClassHint != previous) {
+                int old = fWindowOption ? fWindowOption->layer : WinLayerInvalid;
+                fWindowOption = null;
                 if (fFrame) {
                     fFrame->getFrameHints();
+                    if (fWindowOption && validLayer(fWindowOption->layer))
+                        fFrame->setRequestedLayer(fWindowOption->layer);
+                    else if (validLayer(old))
+                        fFrame->setRequestedLayer(WinLayerNormal);
                     if (taskBarTaskGrouping) {
                         fFrame->removeAppStatus();
                         fFrame->updateAppStatus();
@@ -646,19 +668,7 @@ void YFrameClient::handleProperty(const XPropertyEvent &property) {
 
     case XA_WM_HINTS:
         if (new_prop) prop.wm_hints = true;
-        {
-            Drawable oldPix = iconPixmapHint();
-            Drawable oldMask = iconMaskHint();
-            bool oldUrge = urgencyHint();
-            getWMHints();
-            if (oldPix != iconPixmapHint() || oldMask != iconMaskHint()) {
-                refreshIcon();
-            }
-            if (oldUrge != urgencyHint()) {
-                if (getFrame())
-                    getFrame()->setWmUrgency(urgencyHint());
-            }
-        }
+        updateWMHints();
         prop.wm_hints = new_prop;
         break;
 
@@ -1270,6 +1280,33 @@ Pixmap YFrameClient::iconMaskHint() const {
     return wmHint(IconMaskHint) ? fHints->icon_mask : None;
 }
 
+void YFrameClient::updateWMHints() {
+    Drawable oldPix = iconPixmapHint();
+    Drawable oldMask = iconMaskHint();
+    bool oldUrgency = urgencyHint();
+    getWMHints();
+    if (oldPix != iconPixmapHint() || oldMask != iconMaskHint()) {
+        refreshIcon();
+    }
+    if (fUrgencyTimer) {
+        if (urgencyHint() == false)
+            fUrgencyTimer = null;
+        else if (getFrame() && getFrame()->focused())
+            fHints->flags &= ~XUrgencyHint;
+    }
+    if (oldUrgency != urgencyHint()) {
+        if (getFrame())
+            getFrame()->setWmUrgency(urgencyHint());
+    }
+}
+
+void YFrameClient::clearUrgency() {
+    if (urgencyHint()) {
+        fHints->flags &= ~XUrgencyHint;
+        fUrgencyTimer->setTimer(500L, this, true);
+    }
+}
+
 bool YFrameClient::isDockApp() const {
     return isDockAppIcon() || isDockAppWindow();
 }
@@ -1635,12 +1672,12 @@ bool YFrameClient::getWinTrayHint(int* tray_opt) {
 }
 
 void YFrameClient::setStateHint(int state) {
-    MSG(("set state 0x%8X, saved 0x%8X, win 0x%lx",
-          state, fWinStateHint, handle()));
-
     if (((fWinStateHint ^ state) & WIN_STATE_NET) == 0 || destroyed()) {
         return;
     } else {
+        MSG(("0x%07lx: set state 0x%8X, saved 0x%8X", handle(),
+              state & WIN_STATE_NET, fWinStateHint == InvalidFrameState ?
+              InvalidFrameState : fWinStateHint & WIN_STATE_NET));
         fWinStateHint = state;
     }
 
